######################################################################
#
# EPrints::Utils
#
######################################################################
#
#  This file is part of GNU EPrints 2.
#  
#  Copyright (c) 2000-2004 University of Southampton, UK. SO17 1BJ.
#  
#  EPrints 2 is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#  
#  EPrints 2 is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  
#  You should have received a copy of the GNU General Public License
#  along with EPrints 2; if not, write to the Free Software
#  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
######################################################################


=pod

=head1 NAME

B<EPrints::Utils> - undocumented

=head1 DESCRIPTION

undocumented

=over 4

=cut

######################################################################
#
# INSTANCE VARIABLES:
#
#  $self->{foo}
#     undefined
#
######################################################################

######################################################################
#
#  EPrints Utility module
#
#   Provides various useful functions
#
######################################################################
#
#  This file is part of GNU EPrints 2.
#  
#  Copyright (c) 2000-2004 University of Southampton, UK. SO17 1BJ.
#  
#  EPrints 2 is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#  
#  EPrints 2 is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  
#  You should have received a copy of the GNU General Public License
#  along with EPrints 2; if not, write to the Free Software
#  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
######################################################################

package EPrints::Utils;

use Filesys::DiskSpace;
use Unicode::String qw(utf8 latin1 utf16);
use File::Path;
use URI;
use Carp;

use EPrints::SystemSettings;
use EPrints::XML;

use strict;

$EPrints::Utils::FULLTEXT = "_fulltext_";

my $DF_AVAILABLE;

BEGIN {
	$DF_AVAILABLE = 0;

	sub detect_df 
	{
		my $dir = "/";
		my ($fmt, $res);
	
		# try with statvfs..
		eval 
		{  
			{
				package main;
				require "sys/syscall.ph";
			}
			$fmt = "\0" x 512;
			$res = syscall (&main::SYS_statvfs, $dir, $fmt) ;
			$res == 0;
		}
		# try with statfs..
		|| eval 
		{ 
			{
				package main;
				require "sys/syscall.ph";
			}	
			$fmt = "\0" x 512;
			$res = syscall (&main::SYS_statfs, $dir, $fmt);
			$res == 0;
		}
	}
	unless( $EPrints::SystemSettings::conf->{disable_df} )
	{
		$DF_AVAILABLE = detect_df();
		if( !$DF_AVAILABLE )
		{
			print STDERR <<END;
---------------------------------------------------------------------------
df ("Disk Free" system call) appears to be unavailable on your server. To 
enable it, you should run 'h2ph * */*' (as root) in your /usr/include 
directory. See the EPrints manual for more information.

If you can't get df working on your system, you can work around it by
adding 
  disable_df => 1
to .../eprints2/perl_lib/EPrints/SystemSettings.pm
but you should read the manual about the implications of doing this.
---------------------------------------------------------------------------
END
			exit;
		}
	}
}


######################################################################
# $dirspace = df_dir( $dir );
#
#  Returns the amount of free space in directory $dir, or undef
#  if df could not be used.
# 
######################################################################


######################################################################
=pod

=item EPrints::Utils::df_dir( $dir )

undocumented

=cut
######################################################################

sub df_dir
{
	my( $dir ) = @_;

	return df $dir if( $DF_AVAILABLE );
	die( "Attempt to call df when df function is not available." );
}




######################################################################
=pod

=item EPrints::Utils::render_date( $session, $datevalue )

undocumented

=cut
######################################################################

sub render_date
{
	my( $session, $datevalue ) = @_;

	if( !defined $datevalue )
	{
		return $session->html_phrase( "lib/utils:date_unspecified" );
	}

	# remove 0'd days and months
	$datevalue =~ s/(-0+)+$//;

	my @elements = split /\-/, $datevalue;

	if( !defined $elements[0] || $elements[0] eq "undef" || $elements[0]==0 )
	{
		return $session->html_phrase( "lib/utils:date_unspecified" );
	}

	# 1999
	if( scalar @elements == 1 )
	{
		return $session->make_text( $elements[0] );
	}

	# 1999-02
	if( scalar @elements == 2 )
	{
		return $session->make_text( EPrints::Utils::get_month_label( $session, $elements[1] )." ".$elements[0] );
	}

#	if( $#elements != 2 || $elements[1] < 1 || $elements[1] > 12 )
#	{
#		return $session->html_phrase( "lib/utils:date_invalid" );
#	}

	# 1999-02-02
	return $session->make_text( $elements[2]." ".EPrints::Utils::get_month_label( $session, $elements[1] )." ".$elements[0] );
}


######################################################################
=pod

=item EPrints::Utils::get_month_label( $session, $monthid )

undocumented

=cut
######################################################################

sub get_month_label
{
	my( $session, $monthid ) = @_;

	my $code = sprintf( "lib/utils:month_%02d", $monthid );

	return $session->phrase( $code );
}



######################################################################
=pod

=item $string = EPrints::Utils::make_name_string( $name, [$familylast] )

undocumented

=cut
######################################################################

sub make_name_string
{
	my( $name, $familylast ) = @_;

	my $firstbit = "";
	if( defined $name->{honourific} && $name->{honourific} ne "" )
	{
		$firstbit = $name->{honourific}." ";
	}
	if( defined $name->{given} )
	{
		$firstbit.= $name->{given};
	}
	
	
	my $secondbit = "";
	if( defined $name->{family} )
	{
		$secondbit = $name->{family};
	}
	if( defined $name->{lineage} && $name->{lineage} ne "" )
	{
		$secondbit .= " ".$name->{lineage};
	}

	
	if( defined $familylast && $familylast )
	{
		return $firstbit." ".$secondbit;
	}
	
	return $secondbit.", ".$firstbit;
}

######################################################################
#
# ( $cmp ) = cmp_names( $val_a , $val_b )
#
#  This method compares (alphabetically) two arrays of names. Passed
#  by reference.
#
######################################################################



######################################################################
=pod

=item EPrints::Utils::cmp_namelists( $a, $b, $fieldname )

undocumented

=cut
######################################################################

sub cmp_namelists
{
	my( $a , $b , $fieldname ) = @_;

	my $val_a = $a->get_value( $fieldname );
	my $val_b = $b->get_value( $fieldname );
	return _cmp_names_aux( $val_a, $val_b );
}


######################################################################
=pod

=item EPrints::Utils::cmp_names( $a, $b, $fieldname )

undocumented

=cut
######################################################################

sub cmp_names
{
	my( $a , $b , $fieldname ) = @_;

	my $val_a = $a->get_value( $fieldname );
	my $val_b = $b->get_value( $fieldname );
	return _cmp_names_aux( [$val_a] , [$val_b] );
}

######################################################################
# 
# EPrints::Utils::_cmp_names_aux( $val_a, $val_b )
#
# undocumented
#
######################################################################

sub _cmp_names_aux
{
	my( $val_a, $val_b ) = @_;

	my( $texta , $textb ) = ( "" , "" );
	if( defined $val_a )
	{ 
		foreach( @{$a} ) { $texta.=":$_->{family},$_->{given},$_->{honourific},$_->{lineage}"; } 
	}
	if( defined $val_b )
	{ 
		foreach( @{$b} ) { $textb.=":$_->{family},$_->{given},$_->{honourific},$_->{lineage}"; } 
	}

	return( $texta cmp $textb );
}



######################################################################
=pod

=item EPrints::Utils::cmp_ints( $a, $b, $fieldname )

undocumented

=cut
######################################################################

sub cmp_ints
{
	my( $a , $b , $fieldname ) = @_;
	my $val_a = $a->get_value( $fieldname );
	my $val_b = $b->get_value( $fieldname );
	$val_a = 0 if( !defined $val_a );
	$val_b= 0 if( !defined $val_b);
	return $val_a <=> $val_b
}


######################################################################
=pod

=item EPrints::Utils::cmp_strings( $a, $b, $fieldname )

undocumented

=cut
######################################################################

sub cmp_strings
{
	my( $a , $b , $fieldname ) = @_;
	my $val_a = $a->get_value( $fieldname );
	my $val_b = $b->get_value( $fieldname );
	$val_a = "" if( !defined $val_a );
	$val_b= "" if( !defined $val_b);
	return $val_a cmp $val_b
}


######################################################################
=pod

=item EPrints::Utils::cmp_dates( $a, $b, $fieldname )

undocumented

=cut
######################################################################

sub cmp_dates
{
	my( $a , $b , $fieldname ) = @_;
	return cmp_strings( $a, $b, $fieldname );
}

# replyto / replytoname are optional (both or neither), they set
# the reply-to header.

######################################################################
=pod

=item EPrints::Utils::send_mail( $archive, $langid, $name, $address, $subject, $body, $sig, [$replyto, $replytoname] )

Sends an email. 

$archive - the archive sending the email.
$langid - the language id (eg. "en")
$name - the name of the recipient (UTF-8 encoded string)
$address - the email address of the recipient
$subject - the subject of the message (UTF-8 encoded string)
$body - the body of the message as a DOM tree
$sig - the signature file as a DOM tree
$replyto -  the reply-to email address 
$replytoname - the reply-to name (UTF-8 encoded string)

Returns true if mail sending (appears to have) succeeded. False otherwise.

=cut
######################################################################

sub send_mail
{
	my( $archive, $langid, $name, $address, $subject, $body, $sig, $replyto, $replytoname ) = @_;
	#   Archive   string   utf8   utf8      utf8      DOM    DOM   string    utf8

	my $mail_func = $archive->get_conf( "send_email" );
	if( !defined $mail_func )
	{
		$mail_func = \&send_mail_via_sendmail;
	}

	my $result = &{$mail_func}( $archive, $langid, $name, $address, $subject, $body, $sig, $replyto, $replytoname );

	if( !$result )
	{
		$archive->log( "Failed to send mail.\nTo: $address <$name>\nSubject: $subject\n" );
	}

	return $result;
}

######################################################################
=pod

=item $datestring = EPrints::Utils::email_date()

Returns the current date and time formatted appropriately for an email
Date: field.

=cut
######################################################################

sub email_date()
{
	my @D = qw/Sun Mon Tue Wed Thu Fri Sat/;
	my @M = qw/Jan Feb Mar Apr May Jun Jul Aug Sep Oct Nov Dec/;

	my( $sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst ) = localtime(time);
	my @gmt = gmtime(time);

	my $zone = $hour-$gmt[2];
	if( $zone > 12 ) { $zone-=24; }
	if( $zone < -12 ) { $zone+=24; }

	my $date = $D[$wday].', '.$mday.' '.$M[$mon].' '.($year+1900).' ';
	$date .= sprintf( "%02d:%02d:%02d ", $hour, $min, $sec );
	if( $zone < 0 ) { $date.='-'; $zone = -$zone; } else { $date.='+'; }
	$date .= sprintf( "%02d",$zone ).'00';

	return $date;
}



######################################################################
#
# $encoding = get_encoding($mystring)
# 
# Returns:
# "7-bit" if 7-bit clean
# "utf-8" if utf-8 encoded
# "iso-8859-1" if 8859-1 encoded
# "unknown" if of unknown origin (shouldn't really happen)
#
######################################################################


######################################################################
=pod

=item EPrints::Utils::send_mail_via_smtp( $archive, $langid, $name, $address, $subject, $body, $sig, $replyto, $replytoname )

undocumented

=cut
######################################################################

sub send_mail_via_smtp
{
	my( $archive, $langid, $name, $address, $subject, $body, $sig, $replyto, $replytoname ) = @_;
	#   Archive   string   utf8   utf8      utf8      DOM    DOM   string    utf8

	my $smtphost = $archive->get_conf( 'smtp_server' );

	if( !defined $smtphost )
	{
		$archive->log( "No STMP host has been defined. To fix this, find the full\naddress of your SMTP server (eg. smtp.example.com) and add it\nas the value of smtp_server in\nperl_lib/EPrints/SystemSettings.pm" );
		return( 0 );
	}


	use Net::SMTP;
	my $smtp=Net::SMTP->new( $smtphost );
	if( !defined $smtp )
	{
		$archive->log( "Failed to create smtp connection to $smtphost" );
		return( 0 );
	}
	# test ?
#	unless( open( SENDMAIL, "|".$archive->invocation( "sendmail" ) ) )
#	{
#		$archive->log( "Failed to invoke sendmail: ".
#			$archive->invocation( "sendmail" ) );
#		return( 0 );
#	}

	# Addresses should be 7bit clean, but I'm not checking yet.
	# god only knows what 8bit data does in an email address.

	#cjg should be in the top of the file.
	my $MAILWIDTH = 80;
	my $arcname_q = mime_encode_q( EPrints::Session::best_language( 
		$archive,
		$langid,
		%{$archive->get_conf( "archivename" )} ) );

	my $name_q = mime_encode_q( $name );
	my $subject_q = mime_encode_q( $subject );
	my $adminemail = $archive->get_conf( "adminemail" );

	my $utf8body 	= EPrints::Utils::tree_to_utf8( $body , $MAILWIDTH );
	my $utf8sig	= EPrints::Utils::tree_to_utf8( $sig , $MAILWIDTH );
	my $utf8all	= $utf8body.$utf8sig;
	my $type	= get_encoding($utf8all);
	my $date	= email_date();
	my $content_type_q = 'text/plain; charset="'.$type.'"';
	my $msg = $utf8all;
	if ($type eq "iso-8859-1")
	{
		$msg = $utf8all->latin1; 
	}

	$smtp->mail( $adminemail );
	$smtp->to( $address );
	$smtp->data();

	my $mailheader = <<END;
From: "$arcname_q" <$adminemail>
To: "$name_q" <$address>
Subject: $arcname_q: $subject_q
Date: $date
Precedence: bulk
Content-Type: $content_type_q
Content-Transfer-Encoding: 8bit
END
	if( defined $replyto )
	{
		my $replytoname_q = mime_encode_q( $replytoname );
		$mailheader.= "Reply-To: \"$replytoname_q\" <$replyto>\n";
	}

	# de Unicode::String the strings!
	$smtp->datasend( "$mailheader" );  
	$smtp->datasend( "\n" );
	$smtp->datasend( "$msg" );  
	$smtp->dataend();
	$smtp->quit;

	return( 1 );
}

######################################################################
=pod

=item EPrints::Utils::send_mail_via_sendmail( $archive, $langid, $name, $address, $subject, $body, $sig, $replyto, $replytoname )

undocumented

=cut
######################################################################

sub send_mail_via_sendmail
{
	my( $archive, $langid, $name, $address, $subject, $body, $sig, $replyto, $replytoname ) = @_;
	#   Archive   string   utf8   utf8      utf8      DOM    DOM   string    utf8
	unless( open( SENDMAIL, "|".$archive->invocation( "sendmail" ) ) )
	{
		$archive->log( "Failed to invoke sendmail: ".
			$archive->invocation( "sendmail" ) );
		return( 0 );
	}

	# Addresses should be 7bit clean, but I'm not checking yet.
	# god only knows what 8bit data does in an email address.

	#cjg should be in the top of the file.
	my $MAILWIDTH = 80;
	my $arcname_q = mime_encode_q( EPrints::Session::best_language( 
		$archive,
		$langid,
		%{$archive->get_conf( "archivename" )} ) );

	my $name_q = mime_encode_q( $name );
	my $subject_q = mime_encode_q( $subject );
	my $adminemail = $archive->get_conf( "adminemail" );

	my $utf8body 	= EPrints::Utils::tree_to_utf8( $body , $MAILWIDTH );
	my $utf8sig	= EPrints::Utils::tree_to_utf8( $sig , $MAILWIDTH );
	my $utf8all	= $utf8body.$utf8sig;
	my $type	= get_encoding($utf8all);
	my $date	= email_date();
	my $content_type_q = 'text/plain; charset="'.$type.'"';
	my $msg = $utf8all;
	if ($type eq "iso-8859-1")
	{
		$msg = $utf8all->latin1; 
	}
	#precedence bulk to avoid automail replies?  cjg
	my $mailheader = "";
	if( defined $replyto )
	{
		my $replytoname_q = mime_encode_q( $replytoname );
		$mailheader.= <<END;
Reply-To: "$replytoname_q" <$replyto>
END
	}
	$mailheader.= <<END;
From: "$arcname_q" <$adminemail>
To: "$name_q" <$address>
Subject: $arcname_q: $subject_q
Date: $date
Precedence: bulk
Content-Type: $content_type_q
Content-Transfer-Encoding: 8bit
END

	print SENDMAIL $mailheader;
	print SENDMAIL "\n";
	print SENDMAIL $msg;
	close(SENDMAIL) or return( 0 );
	return( 1 );
}


######################################################################
=pod

=item EPrints::Utils::get_encoding( $string )

undocumented

=cut
######################################################################

sub get_encoding
{
	my( $string ) = @_;

	return "7-bit" if (length($string) == 0);

	my $svnbit = 1;
	my $latin1 = 1;
	my $utf8   = 0;

	foreach($string->unpack())
	{
		$svnbit &= !($_ > 0x79);	
		$latin1 &= !($_ > 0xFF);
		if ($_ > 0xFF)
		{
			$utf8 = 1;	
			last;
		} 
	}
	return "7-bit" if $svnbit;
	return "utf-8" if $utf8;
	return "iso-8859-1" if $latin1;
	return "unknown";
}

# Encode a utf8 string for a MIME header.

######################################################################
=pod

=item EPrints::Utils::mime_encode_q( $string )

undocumented

=cut
######################################################################

sub mime_encode_q
{
	my( $string ) = @_;
	
	my $stringobj = Unicode::String->new();
	$stringobj->utf8( $string );	

	my $encoding = get_encoding($stringobj);

	return $stringobj
		if( $encoding eq "7-bit" );

	return $stringobj
		if( $encoding ne "utf-8" && $encoding ne "iso-8859-1" );

	my @words = split( " ", $stringobj->utf8 );

	foreach( @words )
	{
		my $wordobj = Unicode::String->new();
		$wordobj->utf8( $_ );	
		# don't do words which are 7bit clean
		next if( get_encoding($wordobj) eq "7-bit" );

		my $estr = ( $encoding eq "iso-8859-1" ?
		             $wordobj->latin1 :
			     $wordobj );
		
		$_ = "=?".$encoding."?Q?".encode_str($estr)."?=";
	}

	return join( " ", @words );
}



######################################################################
=pod

=item EPrints::Utils::encode_str( $string )

undocumented

=cut
######################################################################

sub encode_str
{
	my( $string ) = @_;
	my $encoded = "";
        my $i;
        for $i (0..length($string)-1)
        {
                my $o = ord(substr($string,$i,1));
                # less than space, higher or equal than 'DEL' or _ or ?
                if( $o < 0x20 || $o > 0x7E || $o == 0x5F || $o == 0x3F )
                {
                        $encoded.=sprintf( "=%02X", $o );
                }
                else
                {
                        $encoded.=chr($o);
                }
        }
	return $encoded;
}

# ALL cjg get_value should use this.

######################################################################
=pod

=item EPrints::Utils::is_set( $r )

undocumented

=cut
######################################################################

sub is_set
{
	my( $r ) = @_;

	return 0 if( !defined $r );
		
	if( ref($r) eq "" )
	{
		return ($r ne "");
	}
	if( ref($r) eq "ARRAY" )
	{
		foreach( @$r )
		{
			return( 1 ) if( is_set( $_ ) );
		}
		return( 0 );
	}
	if( ref($r) eq "HASH" )
	{
		foreach( keys %$r )
		{
			return( 1 ) if( is_set( $r->{$_} ) );
		}
		return( 0 );
	}
	# Hmm not a scalar, or a hash or array ref.
	# Lets assume it's set. (it is probably a blessed thing)
	return( 1 );
}

# widths smaller than about 3 may totally break, but that's
# a stupid thing to do, anyway.

######################################################################
=pod

=item EPrints::Utils::tree_to_utf8( $node, $width, $pre, $whitespace_before )

undocumented

=cut
######################################################################

sub tree_to_utf8
{
        my( $node, $width, $pre, $whitespace_before ) = @_;

	$whitespace_before = 0 unless defined $whitespace_before;

	unless( EPrints::XML::is_dom( $node ) )
	{
		print STDERR "Oops. tree_to_utf8 got as a node: $node\n";
	}
	if( EPrints::XML::is_dom( $node, "NodeList" ) )
	{
		# Hmm, a node list, not a node.
        	my $string = utf8("");
		my $ws = $whitespace_before;
        	for( my $i=0 ; $i<$node->getLength ; ++$i )
        	{
                	$string .= tree_to_utf8( 
				$node->index( $i ), 
				$width,
 				$pre,
				$ws );
			$ws = _blank_lines( $ws, $string );
		}
		return $string;
	}

        if( defined $width )
        {
                # If we are supposed to be doing an 80 character wide display
                # then only do 78, so the last char does not force a line break.                
		$width = $width - 2;
        }

	
	if( EPrints::XML::is_dom( $node, "Text" ) ||
	    EPrints::XML::is_dom( $node, "CDataSection" ) )
        {
        	my $v = $node->getNodeValue();
                $v =~ s/[\s\r\n\t]+/ /g unless( $pre );
                return $v;
        }

        my $name = $node->getNodeName();

        my $string = utf8("");
	my $ws = $whitespace_before;
        foreach( $node->getChildNodes )
        {
                $string .= tree_to_utf8( 
			$_,
			$width, 
			( $pre || $name eq "pre" || $name eq "mail" ),
			$ws );
		$ws = _blank_lines( $ws, $string );
        }

        if( $name eq "fallback" )
        {
                $string = "*".$string."*";
        }

        # <hr /> only makes sense if we are generating a known width.
        if( $name eq "hr" && defined $width )
        {
                $string = latin1("\n"."-"x$width."\n");
        }

        # Handle wrapping block elements if a width was set.
        if( ( $name eq "p" || $name eq "mail" ) && defined $width)
        {
                my @chars = $string->unpack;
                my @donechars = ();
                my $i;
                while( scalar @chars > 0 )
                {
                        # remove whitespace at the start of a line
                        if( $chars[0] == 32 )
                        {
                                splice( @chars, 0, 1 );
                                next;
                        }

                        # no whitespace at start, so look for first line break
                        $i=0;
                        while( $i<$width && defined $chars[$i] && $chars[$i] !=
10 ) { ++$i; }
                        if( defined $chars[$i] && $chars[$i] == 10 )
                        {
                                push @donechars, splice( @chars, 0, $i+1 );
                                next;
                        }

                        # no line breaks, so if remaining text is smaller
                        # than the width then just add it to the end and
                        # we're done.
                        if( scalar @chars < $width )
                        {
                                push @donechars,@chars;
                                last;
                        }

                        # no line break, more than $width chars.
                        # so look for the last whitespace within $width
                        $i=$width-1;
                        while( $i>=0 && $chars[$i] != 32 ) { --$i; }
                        if( defined $chars[$i] && $chars[$i] == 32 )
                        {
                                # up to BUT NOT INCLUDING the whitespace
                                my @line = splice( @chars, 0, $i );
# This code makes the output "flush" by inserting extra spaces where
# there is currently one. Is that what we want? cjg
#my $j=0;
#while( scalar @line < $width )
#{
#       if( $line[$j] == 32 )
#       {
#               splice(@line,$j,0,-1);
#               ++$j;
#       }
#       ++$j;
#       $j=0 if( $j >= scalar @line );
#}
#foreach(@line) { $_ = 32 if $_ == -1; }
                                push @donechars, @line;

                                # just consume the whitespace
                                splice( @chars, 0, 1);
                                # and a CR...
                                push @donechars,10;
                                next;
                        }

                        # No CR's, no whitespace, just split on width then.
                        push @donechars,splice(@chars,0,$width);

                        # Not the end of the block, so add a \n
                        push @donechars,10;
                }
                $string->pack( @donechars );
        }
	$ws = $whitespace_before;
        if( $name eq "p" )
        {
		while( $ws < 2 ) { $string="\n".$string; ++$ws; }
        }
	$ws = _blank_lines( $whitespace_before, $string );
        if( $name eq "p" )
        {
		while( $ws < 1 ) { $string.="\n"; ++$ws; }
        }
        if( $name eq "br" )
        {
		while( $ws < 1 ) { $string.="\n"; ++$ws; }
        }
        if( $name eq "img" )
        {
		my $alt = $node->getAttribute( "alt" );
		$string = $alt if( defined $alt );
        }
        return $string;
}

sub _blank_lines
{
	my( $n, $str ) = @_;

	$str = "\n"x$n . $str;
	$str =~ s/\[[^\]]*\]//sg;
	$str =~ s/[ 	\r]+//sg;
	my $ws;
	for( $ws = 0; substr( $str, (length $str) - 1 - $ws, 1 ) eq "\n"; ++$ws ) {;}

	return $ws;
}


######################################################################
=pod

=item EPrints::Utils::mkdir( $full_path )

undocumented

=cut
######################################################################

sub mkdir
{
	my( $full_path ) = @_;

	# Make sure $dir is a plain old string (not unicode) as
	# Unicode::String borks mkdir
	$full_path = "$full_path";


	my @created = eval
        {
                return mkpath( $full_path, 0, 0775 );
        };
	if( defined $@ && $@ ne "" ) { warn $@; }
        return ( scalar @created > 0 )
}

# cjg - Potential bug if: <ifset a><ifset b></></> and ifset a is disposed
# then ifset: b is processed it will crash.


######################################################################
=pod

=item EPrints::Utils::render_citation( $obj, $cstyle, $url )

undocumented

=cut
######################################################################

sub render_citation
{
	my( $obj, $cstyle, $url ) = @_;

	# This should belong to the base class of EPrint User Subject and
	# Subscription, if we were better OO people...

	my $session = $obj->get_session;

	my $r= _render_citation_aux( $obj, $session, $cstyle, $url );

	return $r;
}

sub _render_citation_aux
{
	my( $obj, $session, $node, $url ) = @_;
	my $rendered;

	if( EPrints::XML::is_dom( $node, "Text" ) ||
	    EPrints::XML::is_dom( $node, "CDataSection" ) )
	{
		my $rendered = $session->make_doc_fragment;
		my $v = $node->getData;
		my $inside = 0;
		foreach( split( '@' , $v ) )
		{
			if( $inside )
			{
				$inside = 0;
				unless( EPrints::Utils::is_set( $_ ) )
				{
					$rendered->appendChild( 
						$session->make_text( '@' ) );
					next;
				}
                                my $field = EPrints::Utils::field_from_config_string( 
					$obj->get_dataset(), 
					$_ );
				$rendered->appendChild( _citation_field_value( $obj, $field ) );
				next;
			}

			$rendered->appendChild( 
				$session->make_text( $_ ) );
			$inside = 1;
		}
		return $rendered;
	}

	if( EPrints::XML::is_dom( $node, "EntityReference" ) )
	{
		# old style. Deprecated.

		my $fname = $node->getNodeName;
		my $field = $obj->get_dataset()->get_field( $fname );

		return _citation_field_value( $obj, $field );
	}

	my $addkids = $node->hasChildNodes;

	if( EPrints::XML::is_dom( $node, "Element" ) )
	{
		my $name = $node->getTagName;
		$name =~ s/^ep://;

		if( $name eq "ifset" )
		{
			$rendered = $session->make_doc_fragment;
			$addkids = $obj->is_set( $node->getAttribute( "name" ) );
		}
		elsif( $name eq "ifnotset" )
		{
			$rendered = $session->make_doc_fragment;
			$addkids = !$obj->is_set( $node->getAttribute( "name" ) );
		}
		elsif( $name eq "ifmatch" || $name eq "ifnotmatch" )
		{
			my $dataset = $obj->get_dataset;

			my $fieldname = $node->getAttribute( "name" );
			my $merge = $node->getAttribute( "merge" );
			my $value = $node->getAttribute( "value" );
			my $match = $node->getAttribute( "match" );

			my @multiple_names = split /\//, $fieldname;
			my @multiple_fields;
			
			# Put the MetaFields in a list
			foreach (@multiple_names)
			{
				push @multiple_fields, EPrints::Utils::field_from_config_string( $dataset, $_ );
			}
	
			my $sf = EPrints::SearchField->new( 
				$session, 
				$dataset, 
				\@multiple_fields,
				$value,	
				$match,
				$merge );

			$addkids = $sf->get_conditions->item_matches( $obj );

			if( $name eq "ifnotmatch" )
			{
				$addkids = !$addkids;
			}

			$rendered = $session->make_doc_fragment;
		}
		elsif( $name eq "iflink" )
		{
			$rendered = $session->make_doc_fragment;
			$addkids = defined $url;
		}
		elsif( $name eq "ifnotlink" )
		{
			$rendered = $session->make_doc_fragment;
			$addkids = !defined $url;
		}
		elsif( $name eq "linkhere" )
		{
			if( defined $url )
			{
				$rendered = $session->make_element( 
					"a",
					href=>EPrints::Utils::url_escape( 
						$url ) );
			}
			else
			{
				$rendered = $session->make_doc_fragment;
			}
		}
	}

	if( !defined $rendered )
	{
		$rendered = $session->clone_for_me( $node );
	}

	# icky code to spot @title@ in node attributes and replace it.
	my $attrs = $rendered->getAttributes;
	if( $attrs )
	{
		for my $i ( 0..$attrs->getLength-1 )
		{
			my $attr = $attrs->item( $i );
			my $v = $attr->getValue;
			$v =~ s/@([a-z0-9_]+)@/$obj->get_value( $1 )/egi;
			$v =~ s/@@/@/gi;
			$attr->setValue( $v );
		}
	}

	if( $addkids )
	{
		foreach my $child ( $node->getChildNodes )
		{
			$rendered->appendChild(
				_render_citation_aux( 
					$obj,
					$session,
					$child,
					$url ) );			
		}
	}
	return $rendered;
}

sub _citation_field_value
{
	my( $obj, $field ) = @_;

	my $session = $obj->get_session;
	my $fname = $field->get_name;
	my $span = $session->make_element( "span", class=>"field_".$fname );
	my $value = $obj->get_value( $fname );
	$span->appendChild( $field->render_value( 
				$session,
				$value,
				0,
 				1 ) );

	return $span;
}




######################################################################
=pod

=item EPrints::Utils::field_from_config_string( $dataset, $fieldname )

undocumented

=cut
######################################################################

sub field_from_config_string
{
	my( $dataset, $fieldname ) = @_;

	my $modifiers = 0;

	my %q = ();
	if( $fieldname =~ s/^([^;\.]*)(\.id)?(;(.*))?$/$1/ )
	{
		if( defined $4 )
		{
			foreach( split( /;/, $4 ) )
			{
				$q{$_}=1;
				$modifiers = 1;
			}
		}
		if( defined $2 ) 
		{ 
			$q{id} = 1; 
			$modifiers = 1;
		}
	}

	my $field;
	if( $fieldname eq $EPrints::Utils::FULLTEXT )
	{
		$field = new EPrints::MetaField(
				dataset=>$dataset,
				multiple=>1,
				name=>$EPrints::Utils::FULLTEXT,
				type=>"fulltext" );
	}
	else
	{
		$field = $dataset->get_field( $fieldname );
	}

	if( !defined $field )
	{
		EPrints::Config::abort( "Can't make field from config_string: $fieldname" );
	}

	if( $field->get_property( "hasid" ) )
	{
		if( $q{id} )
		{
			$field = $field->get_id_field();
		}
		else
		{
			$field = $field->get_main_field();
		
		}
	}

	unless( $modifiers ) { return $field; }

	$field = $field->clone;

	my $opts = {};
	foreach( keys %q )
	{
		my( $k, $v ) = split( /=/, $_ );
		$v = 1 unless defined $v;
		$opts->{$k} = $v;
	}

	$field->set_property( "render_opts", $opts );
	
	return $field;
}

######################################################################
=pod

=item EPrints::Utils::get_input( $regexp, $prompt, $default )

undocumented

=cut
######################################################################

sub get_input
{
	my( $regexp, $prompt, $default ) = @_;

	$prompt = "" if( !defined $prompt);
	for(;;)
	{
		print $prompt;
		if( defined $default )
		{
			print " [$default] ";
		}
		print "? ";
		my $in = <STDIN>;
		chomp $in;
		if( $in eq "" && defined $default )
		{
			return $default;
		}
		if( $in=~m/^$regexp$/ )
		{
			return $in;
		}
		else
		{
			print "Bad Input, try again.\n";
		}
	}
}


######################################################################
=pod

=item EPrints::Utils::clone( $data )

undocumented

=cut
######################################################################

sub clone
{
	my( $data ) = @_;

	if( ref($data) eq "" )
	{
		return $data;
	}
	if( ref($data) eq "ARRAY" )
	{
		my $r = [];
		foreach( @{$data} )
		{
			push @{$r}, clone( $_ );
		}
		return $r;
	}
	if( ref($data) eq "HASH" )
	{
		my $r = {};
		foreach( keys %{$data} )
		{
			$r->{$_} = clone( $data->{$_} );
		}
		return $r;
	}


	# dunno
	return $data;			
}


######################################################################
=pod

=item EPrints::Utils::crypt_password( $value, $session )

undocumented

=cut
######################################################################

sub crypt_password
{
	my( $value, $session ) = @_;

	return unless EPrints::Utils::is_set( $value );

	my @saltset = ('a'..'z', 'A'..'Z', '0'..'9', '.', '/');
	my $salt = $saltset[time % 64] . $saltset[(time/64)%64];
	my $cryptpass = crypt($value ,$salt);

	return $cryptpass;
}

# Escape everything AFTER the last /

######################################################################
=pod

=item EPrints::Utils::url_escape( $url )

undocumented

=cut
######################################################################

sub url_escape
{
	my( $url ) = @_;

	my $uri = URI->new( $url );
	return $uri->as_string;
}

# Command Version: Prints the GNU style --version comment for a command
# line script. Then exits.

######################################################################
=pod

=item EPrints::Utils::cmd_version( $progname )

undocumented

=cut
######################################################################

sub cmd_version
{
	my( $progname ) = @_;

	my $version_id = $EPrints::SystemSettings::conf->{version_id};
	my $version = $EPrints::SystemSettings::conf->{version};
	
	print <<END;
$progname (GNU EPrints $version_id)
$version

Copyright (C) 2001-2004 University of Southampton

This file is part of GNU EPrints 2.

Copyright (c) 2000-2004 University of Southampton, UK. SO17 1BJ.

EPrints 2 is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

EPrints 2 is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with EPrints 2; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
END
	exit;
}

# This code is for debugging memory leaks in objects.
# It is not used by EPrints except when developing. 
#
# 
# my %OBJARRAY = ();
# my %OBJSCORE = ();
# my %OBJPOS = ();
# my %OBJPOSR = ();
# my $c = 0;


######################################################################
=pod

=item EPrints::Utils::destroy( $ref )

undocumented

=cut
######################################################################

sub destroy
{
	my( $ref ) = @_;
#
#	my $class = delete $OBJARRAY{"$ref"};
#	my $n = delete $OBJPOS{"$ref"};
#	delete $OBJPOSR{$n};
#	
#	$OBJSCORE{$class}--;
#	print "Kill: $ref ($class) [$OBJSCORE{$class}]\n";

}

#my %OBJOLDSCORE = ();
#use Data::Dumper;
#sub debug
#{
#	my @k = sort {$b<=>$a} keys %OBJPOSR;
#	for(0..9)
#	{
#		print "=========================================\n";
#		print $OBJPOSR{$k[$_]}."\n";
#	}
#	foreach( keys %OBJSCORE ) { 
#		my $diff = $OBJSCORE{$_}-$OBJOLDSCORE{$_};
#		if( $diff > 0 ) { $diff ="+$diff"; }
#		print "$_ $OBJSCORE{$_}   $diff\n"; 
#		$OBJOLDSCORE{$_} = $OBJSCORE{$_};
#	}
#}
#
#sub bless
#{
#	my( $ref, $class ) = @_;
#
#	CORE::bless $ref, $class;
#
#	$OBJSCORE{$class}++;
#	print "Make: $ref ($class) [$OBJSCORE{$class}]\n";
#	$OBJARRAY{"$ref"}=$class;
#	$OBJPOS{"$ref"} = $c;
#	#my $x = $ref;
#	$OBJPOSR{$c} = "$c - $ref\n";
#	my $i=1;
#	my @info;
#	while( @info = caller($i++) )
#	{
#		$OBJPOSR{$c}.="$info[3] $info[2]\n";
#	}
#
#
#	if( ref( $ref ) =~ /XML::DOM/  )
#	{// to_string
#		#$OBJPOSR{$c}.= $ref->toString."\n";
#	}
#	++$c;
#
#	return $ref;
#}


######################################################################
=pod

=item $xhtml = EPrints::Utils::render_xhtml_field( $session, $field,
$value )

Return an XHTML DOM object of the contents of $value. In the case of
an error parsing the XML in $value return an XHTML DOM object 
describing the problem.

This is intented to be used by the render_single_value metadata 
field option, as an alternative to the default text renderer. 

This allows through any XML element, so could cause problems if
people start using SCRIPT to make pop-up windows. A later version
may allow a limited set of elements only.

=cut
######################################################################

sub render_xhtml_field
{
	my( $session , $field , $value ) = @_;

	if( !defined $value ) { return $session->make_doc_fragment; }
        my( %c ) = (
                ParseParamEnt => 0,
                ErrorContext => 2,
                NoLWP => 1 );

        my $doc = eval { EPrints::XML::parse_xml_string( "<fragment>".$value."</fragment>" ); };
        if( $@ )
        {
                my $err = $@;
                $err =~ s# at /.*##;
		my $pre = $session->make_element( "pre" );
		$pre->appendChild( $session->make_text( "Error parsing XML: ".$err ) );
		return $pre;
        }
	my $fragment = $session->make_doc_fragment;
	my $top = ($doc->getElementsByTagName( "fragment" ))[0];
	foreach my $node ( $top->getChildNodes )
	{
		$fragment->appendChild(
			$session->clone_for_me( $node, 1 ) );
	}
	EPrints::XML::dispose( $doc );
		
	return $fragment;
}
	

#
# ( $year, $month, $day ) = get_date( $time )
#
#  Static method that returns the given time (in UNIX time, seconds 
#  since 1.1.79) in the format used by EPrints and MySQL (YYYY-MM-DD).
#


######################################################################
=pod

=item EPrints::Utils::get_date( $time )

undocumented

=cut
######################################################################

sub get_date
{
	my( $time ) = @_;

	my @date = localtime( $time );
	my $day = $date[3];
	my $month = $date[4]+1;
	my $year = $date[5]+1900;
	
	# Ensure number of digits
	while( length $day < 2 )
	{
		$day = "0".$day;
	}

	while( length $month < 2 )
	{
		$month = "0".$month;
	}

	return( $year, $month, $day );
}


######################################################################
#
# $datestamp = get_datestamp( $time )
#
#  Static method that returns the given time (in UNIX time, seconds 
#  since 1.1.79) in the format used by EPrints and MySQL (YYYY-MM-DD).
#
######################################################################



######################################################################
=pod

=item EPrints::Utils::get_datestamp( $time )

undocumented

=cut
######################################################################

sub get_datestamp
{
	my( $time ) = @_;

	my( $year, $month, $day ) = EPrints::Utils::get_date( $time );

	return( $year."-".$month."-".$day );
}

######################################################################
=pod

=item $timestamp = EPrints::Utils::get_timestamp()

Return a string discribing the current local date and time.

=cut
######################################################################

sub get_timestamp
{
	my $stamp = "Error in get_timestamp";
	eval {
		use POSIX qw(strftime);
		$stamp = strftime( "%a %b %e %H:%M:%S %Z %Y", localtime);
	};	
	return $stamp;
}

######################################################################
=pod

=item $timestamp = EPrints::Utils::get_UTC_timestamp()

Return a string discribing the current local date and time. 
In UTC Format. eg:

 1957-03-20T20:30:00Z

This the UTC time, not the localtime.

=cut
######################################################################

sub get_UTC_timestamp
{
	my $stamp = "Error in get_UTC_timestamp";
	eval {
		use POSIX qw(strftime);
		$stamp = strftime( "%Y-%m-%dT%H:%M:%SZ", gmtime);
	};

	return $stamp;
}


######################################################################
=pod

=item $boolean = EPrints::Utils::is_in( $needles, $haystack, $matchall )

undocumented

=cut
######################################################################

sub is_in
{
	my( $needles, $haystack, $matchall ) = @_;
	
	if( $matchall )
	{
		foreach my $n ( @{$needles} )
		{
			my $found = 0;
			foreach my $h ( @{$haystack} )
			{
				$found = 1 if( $n eq $h );
			}
			return 0 unless( $found );
		}
		return 1;
	}

	foreach my $n ( @{$needles} )
	{
		foreach my $h ( @{$haystack} )
		{
			return 1 if( $n eq $h );
		}
	}
	return 0;
}

######################################################################
=pod

=item $esc_string = EPrints::Utils::escape_filename( $string )

Take a value and escape it to be a legal filename to go in the /view/
section of the site.

=cut
######################################################################

sub escape_filename
{
	my( $fileid ) = @_;

	return "NULL" if( $fileid eq "" );

	$fileid = utf8( $fileid );

	my $stringobj = Unicode::String->new();
	$stringobj->utf8( $fileid );

	my $hc = [ 48, 49, 50, 51, 52, 53, 54, 55, 56, 57, 65, 66, 67, 68, 69, 70 ];
	
	my @in = $stringobj->unpack;
	my @out = ();
	foreach( @in )
	{
		if( $_ < 33 ) { push @out, 95; next; }
		if( $_ >=48 && $_ <= 57 ) { push @out, $_; next; }
		if( $_ >=65 && $_ <= 90 ) { push @out, $_; next; }
		if( $_ >=97 && $_ <= 122 ) { push @out, $_; next; }
		if( $_ == 44 || $_ == 45 || $_ == 46 || $_ == 58 || $_ == 95 ) { push @out, $_; next; }
		if( $_ < 256 )
		{
			push @out, 61;
			push @out, $hc->[($_ / 16 )%16];
			push @out, $hc->[$_%16];
			next;
		}
		push @out, 61;
		push @out, 61;
		push @out, $hc->[($_ / 0x1000 )%16];
		push @out, $hc->[($_ / 0x100 )%16];
		push @out, $hc->[($_ / 0x10 )%16];
		push @out, $hc->[$_%16];
		
	}
	
	$stringobj->pack( @out );

        return $stringobj;
}

######################################################################
=pod

=item $filesize_text = EPrints::Utils::human_filesize( $size_in_bytes )

Return a human readable version of a filesize. If 0-4095b then show 
as bytes, if 4-4095Kb show as Kb otherwise show as Mb.

eg. Input of 5234 gives "5Kb", input of 3234 gives "3234b".

This is not internationalised, I don't think it needs to be. Let me
know if this is a problem. support@eprints.org

=cut
######################################################################

sub human_filesize
{
	my( $size_in_bytes ) = @_;

	if( $size_in_bytes < 4096 )
	{
		return $size_in_bytes.'b';
	}

	my $size_in_k = int( $size_in_bytes / 1024 );

	if( $size_in_k < 4096 )
	{
		return $size_in_k.'Kb';
	}

	my $size_in_meg = int( $size_in_k / 1024 );

	return $size_in_meg.'Mb';
}

1;

######################################################################
=pod

=back

=cut
